#include "config.h"
#include <avr/io.h>
#include <avr/interrupt.h>
//#include <util/delay.h>
#if UART_TESTOUT
  #include <avr/pgmspace.h>
#endif
#include "mmc.h"

#ifndef SYSCLK
  #define SYSCLK	9600000UL	//Quarz Frequenz in Hz
  //#define SYSCLK  8000000UL
#endif

/*

 Ports auf Eval-Board
  
Fkt.     AT11/12/13   2313     JP40
-----------------------------------
Out      5-PB0 OC0A   17-PB5
MISO     6-PB1 OC0B   18-PB6
SCK      7-PB2        19-PB7
CS       2-PB3        16-PB4
MOSI     3-PB4        15-PB3
  
    Normalize:

    Check Levels before
    ffmpeg -i video.avi -af "volumedetect" -vn -sn -dn -f null /dev/null


    Manual:
    ffmpeg -i input.wav -af "volume=5dB" output.mp3   // Where 5dB is what is missing to 0dB from measurement

    Auto:
    ffmpeg -i input.wav -filter:a loudnorm output.wav
    or
    ffmpeg -i input.wav -filter:a dynaudnorm output.wav

    Manual erzeugte das lauteste Ergebnis

    Zusammenführen:
    ffmpeg -f concat -safe 0 -i mylist.txt -c copy output

    Runtersamplen:
    ffmpeg -i input.mp4 -vn -acodec pcm_u8 -ar 16000 -ac 1 W16K8B.WAV
    
*/

register uint8_t ucBufferIndex asm("r3");  // -12
//uint8_t ucBufferIndex;
register uint8_t ucSampleDur asm("r2");
  #define MAX_SAMPLE_DUR 2  // Hold the same PWM for 4 loops,  3 loops scheint auch zu gehen und hört sich auch besser an.  2 Samples mit 16kHz hört sich noch besser an
#define ucRequest   OCR0B     // use otherwise unused OCR0B to hold flags
//volatile uint8_t ucRequest;
//register uint8_t ucRequest asm("r4");

#if MEASURE_TCNT0_OCWR
  register uint8_t ucTcnt asm("r4");
#endif

uint8_t aucMmcBuf[MAX_RD_SIZE];

ISR (TIM0_OVF_vect)
//ISR (TIM0_COMPB_vect)
{
  if( ++ucSampleDur >= MAX_SAMPLE_DUR )
  {
    uint8_t uc;
    //PINB |= (1<<SPI_MOSI); // Port-Toggle forth on MOSI to check timing on scope
    uc = aucMmcBuf[ucBufferIndex++];
#if MEASURE_TCNT0_OCWR      
    ucTcnt = TCNT0;
#endif      
    OCR0A=uc; 
    //PINB |= (1<<SPI_MOSI); // Port-Toggle back on MOSI
    ucSampleDur=0;
    if( ucBufferIndex == MAX_RD_SIZE ) ucBufferIndex=0;  // Wrap around
    //if( (ucBufferIndex == 0) || (ucBufferIndex== MAX_RD_SIZE/2)  ) // Buffer-Half played out
    if( !(ucBufferIndex & ((MAX_RD_SIZE/2)-1))  ) // Half played out
    {
      ucRequest=1;  // ask main() to get the next chunk
    }
  }
}



#if UART_TESTOUT
void usart_write_char(char c)
{
  if(c == '\n')
    usart_write_char('\r');  // conv \n to \r\n
  //Warten solange bis Zeichen gesendet wurde
  while( !(UCSRA & (1<<UDRE)) );  // wait until transmit-buffer is ready to fetch another byte
  //Ausgabe des Zeichens
  UDR = c;
  return;
}


//Ausgabe eines Strings
/*
void usart_write_str(char *str)
{
  while (*str)
  {
    usart_write_char(*str++);
  }
}
*/

void usart_write_p_str( const char *Buffer )
{
  uint8_t x;

  x = pgm_read_byte(Buffer++);
  while( x )
  {
    usart_write_char(x);
    x = pgm_read_byte(Buffer++);
  }
}


uint8_t ucHexNib2Asc( uint8_t ucVal )
{
  if( ucVal > 9 )
  {
    ucVal+='A'-10;
  }
  else
  {
    ucVal+='0';
  }
  return( ucVal );
}

VOID vSendUartHexByte( uint8_t uc)
{
  usart_write_char(ucHexNib2Asc(   uc>>4  ));
  usart_write_char(ucHexNib2Asc( uc & 0x0f));
  usart_write_char(' ' );
}
#else

  #define  usart_write_p_str( x )
  #define  vSendUartHexByte( uc) 
  #define  usart_write_char( c )
#endif

// minimal-protocoll
// if a protocol-define is used in stream, it will be byte-stuffed with a 0x10

int  __attribute__((OS_main)) main( void )
{
#if MMC_IDLE_MODE  
  uint8_t x;
#endif   
#if MEASURE_TCNT0_OCWR    
  UINT uiTcntAcc;
#endif

  /*
    Initialisierung der Ports
  */
  PORTB = 0x1D;          // no pullups, all outputs high  01 1111
  DDRB  = 0x1D;          // all outputs except PB1 (MISO) 01 1101
  
  /*
    Initialisierung Timer0
  */

  // PWM-Function
  TCCR0A = (1<<COM0A1) | ( 0<<COM0A0)                        | ( 1<<WGM01) | (1<<WGM00); // Clr OC0A-Pin on match = latch in, use Fast PWM-Mode to 0xFF
  TCCR0B = (0<<CS02) | ( 0<<CS01) | (1<<CS00 ) | ( 0<<WGM02);              // 1:1
  //OCR0B = 0 ( default ) Takeover of new OCR0A in ISR ~3µs after rising edge   -> 29 cy late ->
  //OCR0B = -29;       // Looks almost perfect on the scope , try 3cy earlier
  //OCR0B = -32;       // OK, perfect timing
  //OCR0B = 128-32;    // lets see what happens, when we do it at the worst time.  I cant hear any distortions.. 
  // So it seems unnecessary to spend extra code for the perfect position, leave OC0B on reset-default 0, or use it as flags
  
#if UART_TESTOUT
  /*
    Initialisierung UART-Schnittstelle auf 115k2  ( 2313 )
  */
  // Zuerst 115k2
  UCSRA = (1<<U2X); // double speed
  UCSRB = /* (1<<RXEN) |*/ ( 1<<TXEN);   // enable receiver and transmitter PD0 + PD1 go offline
  //UBRRH =  0;  // Reset State already
  UBRRL = 8;  // from Spec for 8MHz / 115200
  //UBRRL = 16;  // from Spec for 8MHz / 57600
#endif  

/* Inits that can be omitted without too much sideeffects  
  OSCCAL = 90; // tune to ca. 10MHz   100 = 11MHz High-Pitched  too slow like this 
  ucBufferIndex = 0;  // PCM-Buffer index in register, if left out, there may be initial scratches 
  ucSampleDur   = 0;  // 
*/  


/*     Say welcome on UART   */
  usart_write_p_str( PSTR( "SD-Read\n" )) ;


#if MEASURE_TCNT0_OCWR    
  uiTcntAcc =0;
#endif    
  
  sei();
  
  while( 1 )
  {
    //_delay_ms( 4000 );
   
#if MMC_NO_RETURN_CODE
    MMC_init(
  #if MMC_IDLE_MODE
              0
  #endif
    );  // Full init to operational
#else
    x =  MMC_init(
  #if MMC_IDLE_MODE
                   0
  #endif
                     );  // Full init to operational
    if( x )
    {
      usart_write_p_str( PSTR("Init Code ") );
      vSendUartHexByte( x );
      continue;  // try again
    }
#endif    

    TIMSK0 |= (1<<TOIE0);  // Enable Interrupt to start playing

    while( 1 )
    {
      if( ucRequest )
      {
        ucRequest=0;  // ACK
        //usart_write_char('#');
        if( ucGetNextChunk() ) // File done now?
        {
          TIMSK0 &= ~(1<<TOIE0);  // disable Interrupt now 
          break;
        }
#if MEASURE_TCNT0_OCWR    
        uiTcntAcc+=ucTcnt;
        if( ++x == 0 )
        {
          vSendUartHexByte( ucTcnt );
        }
#endif    
      }
    }
  }
}
